@@ -3,7 +3,7 @@
 // pyomniFunc.cc              Created on: 2000/06/07
 //                            Author    : Duncan Grisby (dpg1)
 //
-//    Copyright (C) 2002-2008 Apasphere Ltd
+//    Copyright (C) 2002-2013 Apasphere Ltd
 //    Copyright (C) 2000 AT&T Laboratories Cambridge
 //
 //    This file is part of the omniORBpy library
@@ -20,80 +20,12 @@
 //    GNU Lesser General Public License for more details.
 //
 //    You should have received a copy of the GNU Lesser General Public
-//    License along with this library; if not, write to the Free
-//    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
-//    MA 02111-1307, USA
+//    License along with this library. If not, see http://www.gnu.org/licenses/
 //
 //
 // Description:
 //    omniORB API functions
 
-// $Id: pyomniFunc.cc 5868 2009-05-06 16:50:31Z dgrisby $
-// $Log$
-// Revision 1.1.4.9  2008/10/09 15:04:36  dgrisby
-// Python exceptions occurring during unmarshalling were not properly
-// handled. Exception state left set when at traceLevel 0 (thanks
-// Morarenko Kirill).
-//
-// Revision 1.1.4.8  2008/03/24 20:06:30  dgrisby
-// Memory leak in minor code stringification.
-//
-// Revision 1.1.4.7  2006/01/17 17:38:21  dgrisby
-// Expose omniORB.setClientConnectTimeout function.
-//
-// Revision 1.1.4.6  2005/11/09 12:33:32  dgrisby
-// Support POA LocalObjects.
-//
-// Revision 1.1.4.5  2005/07/22 17:41:07  dgrisby
-// Update from omnipy2_develop.
-//
-// Revision 1.1.4.4  2005/06/24 17:36:00  dgrisby
-// Support for receiving valuetypes inside Anys; relax requirement for
-// old style classes in a lot of places.
-//
-// Revision 1.1.4.3  2005/04/14 13:50:59  dgrisby
-// New traceTime, traceInvocationReturns functions; removal of omniORB::logf.
-//
-// Revision 1.1.4.2  2005/01/07 00:22:33  dgrisby
-// Big merge from omnipy2_develop.
-//
-// Revision 1.1.4.1  2003/03/23 21:51:56  dgrisby
-// New omnipy3_develop branch.
-//
-// Revision 1.1.2.11  2002/11/27 00:18:25  dgrisby
-// Per thread / per objref timeouts.
-//
-// Revision 1.1.2.10  2002/08/16 19:27:36  dgrisby
-// Documentation update. Minor ORB updates to match docs.
-//
-// Revision 1.1.2.9  2002/05/26 00:56:22  dgrisby
-// traceInvocations() function.
-//
-// Revision 1.1.2.8  2002/02/25 15:33:50  dpg1
-// Expose omniORB logger to Python.
-//
-// Revision 1.1.2.7  2001/09/24 10:48:28  dpg1
-// Meaningful minor codes.
-//
-// Revision 1.1.2.6  2001/08/21 12:48:27  dpg1
-// Meaningful exception minor code strings.
-//
-// Revision 1.1.2.5  2001/08/21 10:52:42  dpg1
-// Update to new ORB core APIs.
-//
-// Revision 1.1.2.4  2001/04/09 15:22:16  dpg1
-// Fixed point support.
-//
-// Revision 1.1.2.3  2000/11/22 14:43:58  dpg1
-// Support code set conversion and wchar/wstring.
-//
-// Revision 1.1.2.2  2000/11/06 17:07:17  dpg1
-// maxTcpConnectionPerServer function
-//
-// Revision 1.1.2.1  2000/10/13 13:55:28  dpg1
-// Initial support for omniORB 4.
-//
-
 #include <omnipy.h>
 #include <pyThreadCache.h>
 #include <omniORB4/giopEndpoint.h>
@@ -103,12 +35,17 @@ OMNI_USING_NAMESPACE(omni);
 
 
 static PyObject* transientEHtuple   = 0;
+static PyObject* timeoutEHtuple     = 0;
 static PyObject* commFailureEHtuple = 0;
 static PyObject* systemEHtuple      = 0;
 
 
-static CORBA::Boolean transientEH(void* cookie, CORBA::ULong retries,
-				  const CORBA::TRANSIENT& ex)
+static inline CORBA::Boolean
+exceptionHandler(void*                         cookie,
+                 PyObject*                     def_eh,
+                 CORBA::ULong                  retries,
+                 const CORBA::SystemException& ex,
+                 const char*                   kind)
 {
   PyObject* tuple;
 
@@ -117,130 +54,83 @@ static CORBA::Boolean transientEH(void* cookie, CORBA::ULong retries,
   else
     tuple = transientEHtuple;
 
-  OMNIORB_ASSERT(PyTuple_Check(tuple));
-
-  PyObject* pyfn     = PyTuple_GET_ITEM(tuple, 0);
-  PyObject* pycookie = PyTuple_GET_ITEM(tuple, 1);
-  PyObject* r;
-  CORBA::Boolean ret;
-
   {
     omnipyThreadCache::lock _t;
-    PyObject* pyex = omniPy::createPySystemException(ex);
-    r = PyObject_CallFunction(pyfn, (char*)"OiN", pycookie, retries, pyex);
+
+    OMNIORB_ASSERT(PyTuple_Check(tuple));
+
+    PyObject* pyfn     = PyTuple_GET_ITEM(tuple, 0);
+    PyObject* pycookie = PyTuple_GET_ITEM(tuple, 1);
+    PyObject* pyex     = omniPy::createPySystemException(ex);
+
+    omniPy::PyRefHolder r(PyObject_CallFunction(pyfn, (char*)"OiN",
+                                                pycookie, retries, pyex));
     
-    if (!r) {
+    if (!r.valid()) {
       if (omniORB::trace(1)) {
-	omniORB::logs(1, "Python TRANSIENT exception handler failed. "
-		      "Traceback follows:");
+        {
+          omniORB::logger log;
+          log << "Python " << kind
+              << "  exception handler failed. Traceback follows:\n";
+        }
 	PyErr_Print();
       }
       else
 	PyErr_Clear();
+
       return 0;
     }
-    if (!PyInt_Check(r)) {
-      omniORB::logs(1, "Python TRANSIENT exception handler returned an "
-		    "invalid object.");
-      Py_DECREF(r);
+    int ret = PyObject_IsTrue(r);
+
+    if (ret == -1) {
+      if (omniORB::trace(1)) {
+        {
+          omniORB::logger log;
+          log << "Python " << kind
+              << "  exception handler returned an invalid object.\n";
+        }
+      }
+      else
+	PyErr_Clear();
+
       return 0;
     }
-    ret = PyInt_AS_LONG(r);
-    Py_DECREF(r);
+    return ret;
   }
-  return ret;
 }
 
 
+static CORBA::Boolean transientEH(void* cookie, CORBA::ULong retries,
+				  const CORBA::TRANSIENT& ex)
+{
+  return exceptionHandler(cookie, transientEHtuple, retries, ex, "TRANSIENT");
+}
+
+static CORBA::Boolean timeoutEH(void* cookie, CORBA::ULong retries,
+				const CORBA::TIMEOUT& ex)
+{
+  return exceptionHandler(cookie, timeoutEHtuple, retries, ex, "TIMEOUT");
+}
+
 static CORBA::Boolean commFailureEH(void* cookie, CORBA::ULong retries,
 				    const CORBA::COMM_FAILURE& ex)
 {
-  PyObject* tuple;
-
-  if (cookie)
-    tuple = (PyObject*)cookie;
-  else
-    tuple = commFailureEHtuple;
-
-  OMNIORB_ASSERT(PyTuple_Check(tuple));
-
-  PyObject* pyfn     = PyTuple_GET_ITEM(tuple, 0);
-  PyObject* pycookie = PyTuple_GET_ITEM(tuple, 1);
-  PyObject* r;
-  CORBA::Boolean ret;
-
-  {
-    omnipyThreadCache::lock _t;
-    PyObject* pyex = omniPy::createPySystemException(ex);
-    r = PyObject_CallFunction(pyfn, (char*)"OiN", pycookie, retries, pyex);
-    
-    if (!r) {
-      if (omniORB::trace(1)) {
-	omniORB::logs(1, "Python COMM_FAILURE exception handler failed. "
-		      "Traceback follows:");
-	PyErr_Print();
-      }
-      else
-	PyErr_Clear();
-      return 0;
-    }
-    if (!PyInt_Check(r)) {
-      omniORB::logs(1, "Python COMM_FAILURE exception handler returned an "
-		    "invalid object.");
-      Py_DECREF(r);
-      return 0;
-    }
-    ret = PyInt_AS_LONG(r);
-    Py_DECREF(r);
-  }
-  return ret;
+  return exceptionHandler(cookie, commFailureEHtuple, retries,
+                          ex, "COMM_FAILURE");
 }
 
 static CORBA::Boolean systemEH(void* cookie, CORBA::ULong retries,
 			       const CORBA::SystemException& ex)
 {
-  PyObject* tuple;
-
-  if (cookie)
-    tuple = (PyObject*)cookie;
-  else
-    tuple = systemEHtuple;
-
-  OMNIORB_ASSERT(PyTuple_Check(tuple));
-
-  PyObject* pyfn     = PyTuple_GET_ITEM(tuple, 0);
-  PyObject* pycookie = PyTuple_GET_ITEM(tuple, 1);
-  PyObject* r;
-  CORBA::Boolean ret;
-
-  {
-    omnipyThreadCache::lock _t;
-    PyObject* pyex = omniPy::createPySystemException(ex);
-    r = PyObject_CallFunction(pyfn, (char*)"OiN", pycookie, retries, pyex);
-    
-    if (!r) {
-      if (omniORB::trace(1)) {
-	omniORB::logs(1, "Python SystemException handler failed. "
-		      "Traceback follows:");
-	PyErr_Print();
-      }
-      else
-	PyErr_Clear();
-      return 0;
-    }
-    if (!PyInt_Check(r)) {
-      omniORB::logs(1, "Python SystemException handler returned an "
-		    "invalid object.");
-      Py_DECREF(r);
-      return 0;
-    }
-    ret = PyInt_AS_LONG(r);
-    Py_DECREF(r);
-  }
-  return ret;
+  return exceptionHandler(cookie, systemEHtuple, retries,
+                          ex, "SystemException");
 }
 
+
 extern "C" {
+
+#if (PY_VERSION_HEX < 0x03000000)
+
   static void removeDummyOmniThread(void* vself) {
     if ((omni_thread*)vself == omni_thread::self()) {
       omniORB::logs(10, "Remove dummy omni thread.");
@@ -250,6 +140,21 @@ extern "C" {
     else
       omniORB::logs(5, "Unable to release dummy omni_thread.");
   }
+
+#else
+
+  static void removeDummyOmniThread(PyObject* cap) {
+    void* vself = PyCapsule_GetPointer(cap, 0);
+
+    if ((omni_thread*)vself == omni_thread::self()) {
+      omniORB::logs(10, "Remove dummy omni thread.");
+      omniPy::InterpreterUnlocker _u;
+      omni_thread::release_dummy();
+    }
+    else
+      omniORB::logs(5, "Unable to release dummy omni_thread.");
+  }
+#endif
 }
 
 
@@ -284,7 +189,12 @@ omniPy::ensureOmniThread()
 
   // Create a CObject with a suitable destructor function and set it
   // as an attribute of the current thread.
+#if (PY_VERSION_HEX < 0x03000000)
   PyObject* cobj = PyCObject_FromVoidPtr(self, removeDummyOmniThread);
+#else
+  PyObject* cobj = PyCapsule_New(self, 0, removeDummyOmniThread);
+#endif
+
   PyObject_SetAttrString(current, (char*)"__omni_thread", cobj);
 
   // Use an evil hack to make sure the __omni_thread member is
@@ -306,7 +216,6 @@ omniPy::ensureOmniThread()
 }
 
 
-
 extern "C" {
 
   static char transient_doc [] =
@@ -335,8 +244,7 @@ extern "C" {
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
-      CORBA::Object_ptr objref =
-	(CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
+      CORBA::Object_ptr objref = omniPy::getObjRef(pyobjref);
 
       RAISE_PY_BAD_PARAM_IF(!objref, BAD_PARAM_WrongPythonType);
 
@@ -356,6 +264,52 @@ extern "C" {
     return Py_None;
   }
 
+  static char timeout_doc [] =
+  "installTimeoutExceptionHandler(cookie, function [, object]) -> None\n"
+  "\n"
+  "Install a handler for TIMEOUT exceptions, for all objects or a\n"
+  "specified object. The function must have the signature:\n"
+  "\n"
+  "  function(cookie, retries, exc) -> boolean\n"
+  "\n"
+  "where cookie is the object passed on installation, retries is the\n"
+  "number of times this operation invocation has been retried, and exc is\n"
+  "the TIMEOUT exception object. If the function returns TRUE, the\n"
+  "operation invocation is retried; if it returns FALSE, the TIMEOUT\n"
+  "exception is thrown to the application.\n";
+
+  static PyObject* pyomni_installTimeoutExceptionHandler(PyObject* self,
+							   PyObject* args)
+  {
+    PyObject *pycookie, *pyfn, *pyobjref = 0;
+
+    if (!PyArg_ParseTuple(args, (char*)"OO|O", &pycookie, &pyfn, &pyobjref))
+      return 0;
+
+
+    RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
+
+    if (pyobjref) {
+      CORBA::Object_ptr objref = omniPy::getObjRef(pyobjref);
+
+      RAISE_PY_BAD_PARAM_IF(!objref, BAD_PARAM_WrongPythonType);
+
+      PyObject* tuple = Py_BuildValue((char*)"OO", pyfn, pycookie);
+      PyObject_SetAttrString(pyobjref, (char*)"__omni_timeout", tuple);
+      omniORB::installTimeoutExceptionHandler(objref, (void*)tuple,
+					      timeoutEH);
+    }
+    else {
+      Py_XDECREF(timeoutEHtuple);
+      timeoutEHtuple = Py_BuildValue((char*)"OO", pyfn, pycookie);
+      OMNIORB_ASSERT(timeoutEHtuple);
+      omniORB::installTimeoutExceptionHandler((void*)timeoutEHtuple,
+					      timeoutEH);
+    }
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
+
   static char commfailure_doc [] =
   "installCommFailureExceptionHandler(cookie, function [, object]) -> None\n"
   "\n"
@@ -381,8 +335,7 @@ extern "C" {
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
-      CORBA::Object_ptr objref =
-	(CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
+      CORBA::Object_ptr objref = omniPy::getObjRef(pyobjref);
 
       RAISE_PY_BAD_PARAM_IF(!objref, BAD_PARAM_WrongPythonType);
 
@@ -428,8 +381,7 @@ extern "C" {
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
-      CORBA::Object_ptr objref =
-	(CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
+      CORBA::Object_ptr objref = omniPy::getObjRef(pyobjref);
 
       RAISE_PY_BAD_PARAM_IF(!objref, BAD_PARAM_WrongPythonType);
 
@@ -458,13 +410,13 @@ extern "C" {
   static PyObject* pyomni_traceLevel(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceLevel);
+      return Int_FromLong(omniORB::traceLevel);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceLevel = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceLevel = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -483,13 +435,13 @@ extern "C" {
   static PyObject* pyomni_traceExceptions(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceExceptions);
+      return Int_FromLong(omniORB::traceExceptions);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceExceptions = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceExceptions = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -508,13 +460,13 @@ extern "C" {
   static PyObject* pyomni_traceInvocations(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceInvocations);
+      return Int_FromLong(omniORB::traceInvocations);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceInvocations = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceInvocations = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -533,13 +485,13 @@ extern "C" {
   static PyObject* pyomni_traceInvocationReturns(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceInvocationReturns);
+      return Int_FromLong(omniORB::traceInvocationReturns);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceInvocationReturns = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceInvocationReturns = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -558,13 +510,13 @@ extern "C" {
   static PyObject* pyomni_traceThreadId(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceThreadId);
+      return Int_FromLong(omniORB::traceThreadId);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceThreadId = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceThreadId = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -583,13 +535,13 @@ extern "C" {
   static PyObject* pyomni_traceTime(PyObject* self, PyObject* args)
   {
     if (PyTuple_GET_SIZE(args) == 0) {
-      return PyInt_FromLong(omniORB::traceTime);
+      return Int_FromLong(omniORB::traceTime);
     }
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pytl = PyTuple_GET_ITEM(args, 0);
 
-      if (PyInt_Check(pytl)) {
-	omniORB::traceTime = PyInt_AS_LONG(pytl);
+      if (Int_Check(pytl)) {
+	omniORB::traceTime = Int_AS_LONG(pytl);
 	Py_INCREF(Py_None);
 	return Py_None;
       }
@@ -634,7 +586,7 @@ extern "C" {
 
       if (ncs_c) {
 	const char* ncs = ncs_c->name();
-	return PyString_FromString((char*)ncs);
+	return String_FromString((char*)ncs);
       }
       else {
 	Py_INCREF(Py_None);
@@ -644,10 +596,15 @@ extern "C" {
     else if (PyTuple_GET_SIZE(args) == 1) {
       PyObject* pyncs = PyTuple_GET_ITEM(args, 0);
 
-      if (PyString_Check(pyncs)) {
+      if (String_Check(pyncs)) {
 	try {
 	  omniCodeSet::NCS_C* ncs_c;
-	  ncs_c = omniCodeSet::getNCS_C(PyString_AS_STRING(pyncs));
+	  ncs_c = omniCodeSet::getNCS_C(String_AS_STRING(pyncs));
+
+          if (!ncs_c)
+            OMNIORB_THROW(NO_RESOURCES, NO_RESOURCES_CodeSetNotSupported,
+                          CORBA::COMPLETED_NO);
+
 	  orbParameters::nativeCharCodeSet = ncs_c;
 	}
 	OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
@@ -711,17 +668,20 @@ extern "C" {
     if (!pyminor)
       PyErr_Clear();
 
-    if (!(pyrepoId && PyString_Check(pyrepoId) &&
-	  pyminor  && (PyInt_Check(pyminor) || PyLong_Check(pyminor)))) {
+    if (!(pyrepoId && String_Check(pyrepoId) &&
+	  pyminor  && (Int_Check(pyminor) || PyLong_Check(pyminor)))) {
       Py_INCREF(Py_None);
       return Py_None;
     }
-    const char* repoId = PyString_AS_STRING(pyrepoId);
+    const char* repoId = String_AS_STRING(pyrepoId);
 
     CORBA::ULong minor;
+#if (PY_VERSION_HEX < 0x03000000)
     if (PyInt_Check(pyminor))
       minor = PyInt_AS_LONG(pyminor);
-    else {
+    else
+#endif
+      {
       minor = PyLong_AsUnsignedLong(pyminor);
       if (minor == (CORBA::ULong)-1 && PyErr_Occurred())
 	PyErr_Clear();
@@ -741,7 +701,7 @@ extern "C" {
 #undef ToStringIfMatch
 
     if (str)
-      return PyString_FromString(str);
+      return String_FromString(str);
     else {
       Py_INCREF(Py_None);
       return Py_None;
@@ -769,8 +729,7 @@ extern "C" {
       if (!PyArg_ParseTuple(args, (char*)"Oi", &pyobjref, &timeout))
 	return 0;
 
-      CORBA::Object_ptr objref =
-	(CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
+      CORBA::Object_ptr objref = omniPy::getObjRef(pyobjref);
 
       RAISE_PY_BAD_PARAM_IF(!objref, BAD_PARAM_WrongPythonType);
 
@@ -802,6 +761,32 @@ extern "C" {
     return Py_None;
   }
 
+  static char setClientThreadCallDeadline_doc [] =
+  "setClientThreadCallDeadline(secs)\n"
+  "\n"
+  "Set the client call deadline for the calling thread.\n";
+
+  static PyObject* pyomni_setClientThreadCallDeadline(PyObject* self,
+                                                      PyObject* args)
+  {
+    double deadline;
+    if (!PyArg_ParseTuple(args, (char*)"d", &deadline))
+      return 0;
+
+    unsigned long s, ns;
+    s  = (unsigned long)deadline;
+    ns = (deadline - s) * 1000000000.0;
+
+    try {
+      omniPy::ensureOmniThread();
+      omniORB::setClientThreadCallDeadline(s, ns);
+    }
+    OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
+
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
+
   static char setClientConnectTimeout_doc [] =
   "setClientConnectTimeout(millisecs)\n"
   "\n"
@@ -840,7 +825,7 @@ extern "C" {
     int j;
 
     for (i = ifaddrs->begin(), j=0; i != ifaddrs->end(); i++, j++) {
-      PyList_SetItem(pyaddrs, j, PyString_FromString(*i));
+      PyList_SetItem(pyaddrs, j, String_FromString(*i));
     }
 
     return pyaddrs;
@@ -856,8 +841,8 @@ extern "C" {
   static PyObject* pyomni_setPersistentServerIdentifier(PyObject* self,
 							PyObject* args)
   {
-    char* idstr;
-    int   idlen;
+    char*      idstr;
+    Py_ssize_t idlen;
 
     if (!PyArg_ParseTuple(args, (char*)"s#", &idstr, &idlen))
       return 0;
@@ -886,14 +871,10 @@ extern "C" {
     if (!PyArg_ParseTuple(args, (char*)"OO", &pyold, &pynew))
       return 0;
 
-    CORBA::Object_ptr oldobj =
-      (CORBA::Object_ptr)omniPy::getTwin(pyold, OBJREF_TWIN);
+    CORBA::Object_ptr oldobj = omniPy::getObjRef(pyold);
+    CORBA::Object_ptr newobj = omniPy::getObjRef(pynew);
 
     RAISE_PY_BAD_PARAM_IF(!oldobj, BAD_PARAM_WrongPythonType);
-
-    CORBA::Object_ptr newobj =
-      (CORBA::Object_ptr)omniPy::getTwin(pynew, OBJREF_TWIN);
-
     RAISE_PY_BAD_PARAM_IF(!newobj, BAD_PARAM_WrongPythonType);
 
     omni::locationForward(oldobj->_PR_getobj(), newobj->_PR_getobj(), 0);
@@ -907,6 +888,10 @@ extern "C" {
      pyomni_installTransientExceptionHandler,
      METH_VARARGS, transient_doc},
 
+    {(char*)"installTimeoutExceptionHandler",
+     pyomni_installTimeoutExceptionHandler,
+     METH_VARARGS, timeout_doc},
+
     {(char*)"installCommFailureExceptionHandler",
      pyomni_installCommFailureExceptionHandler,
      METH_VARARGS, commfailure_doc},
@@ -963,6 +948,10 @@ extern "C" {
      pyomni_setClientThreadCallTimeout,
      METH_VARARGS, setClientThreadCallTimeout_doc},
 
+    {(char*)"setClientThreadCallDeadline",
+     pyomni_setClientThreadCallDeadline,
+     METH_VARARGS, setClientThreadCallDeadline_doc},
+
     {(char*)"setClientConnectTimeout",
      pyomni_setClientConnectTimeout,
      METH_VARARGS, setClientConnectTimeout_doc},
@@ -983,9 +972,37 @@ extern "C" {
   };
 }
 
+#if (PY_VERSION_HEX < 0x03000000)
+
 void
 omniPy::initomniFunc(PyObject* d)
 {
   PyObject* m = Py_InitModule((char*)"_omnipy.omni_func", pyomni_methods);
   PyDict_SetItemString(d, (char*)"omni_func", m);
 }
+
+#else
+
+ static struct PyModuleDef omni_func_module = {
+   PyModuleDef_HEAD_INIT,
+   "_omnipy.omni_func",
+   "omniORB API",
+   -1,
+   pyomni_methods,
+   NULL,
+   NULL,
+   NULL,
+   NULL
+ };
+
+void
+omniPy::initomniFunc(PyObject* d)
+{
+  PyObject* m = PyModule_Create(&omni_func_module);
+  if (!m)
+    return;
+
+  PyDict_SetItemString(d, (char*)"omni_func", m);
+}
+
+#endif
